import datetime
import json
import zlib

from peewee import BlobField


class JSONGzipField(BlobField):
    def python_value(self, value):
        if value is not None:
            if isinstance(value, memoryview):
                value = bytes(value)
            return json.loads(zlib.decompress(value))

    def db_value(self, value):
        if value is not None:
            encoded = json.dumps(value, cls=DateTimeJSONEncoder).encode("utf-8")
            return self._constructor(zlib.compress(encoded))


class DateTimeJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            obj = obj.replace(microsecond=0, tzinfo=None)
            return obj.isoformat()
        if isinstance(obj, datetime.date):
            return obj.isoformat()
        else:
            return super(DateTimeJSONEncoder, self).default(obj)



class Issue:

    DOC_FIELDS = (
        "key", "assignee", "reporter", "status", "summary", "description", "comments_str",
    )
    KEYWORD_FIELDS = ("labels", "components")

    def __init__(self, raw_data):
        self.data = raw_data

    @property
    def key(self):
        return self.data["key"]["#text"]

    @property
    def assignee(self):
        return str(self.data["assignee"]["@username"])

    @property
    def reporter(self):
        return self.data["reporter"]["@username"]

    @property
    def status(self):
        return self.data["status"]["#text"]

    @property
    def summary(self):
        return self.data["summary"]

    @property
    def description(self):
        return self.data["description"]

    @property
    def labels(self):
        labels = self.data.get("labels")
        if not labels:
            return []
        labels = labels["label"]
        if isinstance(labels, dict):
            labels = [labels]
        return labels

    @property
    def components(self):
        components = self.data.get("component")
        if not components:
            return []
        if isinstance(components, str):
            components = [components]
        return components

    @property
    def comments_str(self):
        comments = self.data.get("comments")
        if not comments:
            return ""
        if isinstance(comments["comment"], dict):
            comments = [comments["comment"]]
        else:
            comments = comments["comment"]
        result = []
        for comment in comments:
            result.append("{} {}: {}".format(comment["@created"], comment["@author"], comment["#text"]))
        return "\n".join(result)

    def to_doc(self):
        doc = {field: getattr(self, field) for field in self.DOC_FIELDS}
        for field in self.KEYWORD_FIELDS:
            doc[field] = " ".join(getattr(self, field))
        return doc

    def to_string(self, with_description=True, with_comments=True):
        result = [
            "{key} ({status}) [{reporter} -> {assignee}] {summary}".format(
                key=self.key, status=self.status, reporter=self.reporter, assignee=self.assignee,
                summary=self.summary,
            )]
        if with_description:
            result.append(self.description or "")
        if with_comments:
            result.append(self.comments_str)
        return "\n".join(result)



import argparse
import os
import re

from jirafts.searcher import JiraDownloader, JiraSearcher
from jirafts import db


def parse_args():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "--index", help="Index dir location",
        default=os.path.join(os.path.expanduser("~"), ".jirafts", "default_index")
    )

    output = argparse.ArgumentParser(add_help=False)

    output.add_argument("--html", help="Output HTML", default=False, action="store_true")
    output.add_argument(
        "-s", "--short", help="Short output (no description)", default=False, action="store_true"
    )
    output.add_argument(
        "-c", "--comments", help="Show comments", default=False, action="store_true"
    )

    subparsers = parser.add_subparsers(dest="action")
    subparsers.required = True
    sync = subparsers.add_parser("sync")
    search = subparsers.add_parser("search", parents=[output])
    dump = subparsers.add_parser("dump", parents=[output])
    grep = subparsers.add_parser("grep", parents=[output])

    sync.add_argument("-u", "--url", help="JIRA url", required=True)
    sync.add_argument("-t", "--auth", "--token", help="JIRA password or API token", default=None)
    sync.add_argument("-a", "--all", help="Sync all. Not only recently updated", default=False, action="store_true")
    sync.add_argument("-p", "--project", help="Projects to sync. All by default", default=[], nargs="*")
    sync.add_argument("-o", "--optimize", help="Optimize index", default=False, action="store_true")
    sync.add_argument("-c", "--concurrency", help="Number of parallel downloads", default=8, type=int)

    search.add_argument("-l", "--limit", help="Limit results", default=10, type=int)
    search.add_argument(
        "--no-hl", help="Disable highlighting of matched terms", default=False, action="store_true"
    )
    search.add_argument("text")
    grep.add_argument("regex")
    grep.add_argument("-i", "--ignore-case", help="Ignore case", default=False, action="store_true")

    res = parser.parse_args()
    return res


def _get_token_from_file(token_filename):
    with open(token_filename) as fp:
        return fp.read().strip()


def main():
    args = parse_args()
    os.makedirs(args.index, exist_ok=True)
    db_path = os.path.join(args.index, "jira.db")
    db.connect(db_path)
    searcher = JiraSearcher(index_dir=args.index)

    if args.action == "sync":
        auth = args.auth
        if auth:
            if os.path.exists(auth):
                auth = _get_token_from_file(auth)
            auth = tuple(auth.split(":", 1))
        downloader = JiraDownloader(url=args.url, auth=auth, concurrency=args.concurrency)
        searcher.update_db(downloader, projects=args.project, all_issues=args.all, optimize=args.optimize)
    else:
        short = args.short
        comments = not short and args.comments
        if args.action == "dump":
            searcher.dump(html=args.html, comments=comments, description=not args.short)
        elif args.action == "search":
            searcher.search(
                args.text, limit=args.limit, html=args.html,
                comments=comments, description=not short, search_comments=comments,
                highlight=not args.no_hl,
            )
        elif args.action == "grep":
            flags = 0
            if args.ignore_case:
                flags |= re.IGNORECASE
            regex = re.compile("(?P<word>{})".format(args.regex), flags)
            searcher.dump(
                html=args.html, comments=comments, description=not args.short,
                regex=regex, regex_group_name="word",
            )



from whoosh import highlight


def ansi_highlight(s):
    return "\033[92m{}\033[0m".format(s)


class AnsiColorFormatter(highlight.Formatter):
    def format_token(self, text, token, replace=False):
        token_text = highlight.get_text(text, token, replace)
        return ansi_highlight(token_text)



import datetime
import os
import concurrent.futures
import posixpath

if os.getenv("JIRAFTS_DEBUG", False):
    import requests_cache
    requests_cache.install_cache()
    print("Using cache for HTTP queries")
import requests
import xmltodict


class JiraDownloader:
    DATE_FORMAT =  "%a, %d %b %Y %H:%M:%S %z"
    SEARCH_DATE_FORMAT = "%Y-%m-%d %H:%M"
    DATE_FIELDS = ["updated", "created"]

    def __init__(self, url, auth=None, concurrency=8):
        self._client = requests.Session()
        self._auth = auth
        self._url = url
        self._concurrency = concurrency

    def iter_issues(self, projects=None, limit=None, per_page=50, min_date=None, max_date=None):
        start = 0
        if limit:
            per_page = min(per_page, limit)
        total = 0
        query_parts = []
        if projects:
            query_parts.append("project in ({})".format(", ".join(projects)))
        if min_date and max_date:
            query_parts.append(
                "NOT (updated > '{min_date}' AND updated < '{max_date}')".format(
                    min_date=min_date.strftime(self.SEARCH_DATE_FORMAT),
                    max_date=max_date.strftime(self.SEARCH_DATE_FORMAT),
                )
            )

        query = "{} ORDER BY updated DESC".format(" AND ".join(query_parts))
        data = self._request_issues(query=query, start=start, per_page=per_page)
        total_results = int(data["issue"]["@total"])
        if not total_results:
            print("No results downloaded. Wrong query or auth data")
            return

        futures = [concurrent.futures.Future()]
        futures[0].set_result(data)
        if not limit:
            limit = total_results
        elif limit > total_results:
            limit = total_results

        with concurrent.futures.ThreadPoolExecutor(self._concurrency) as tp:
            for start in range(per_page, limit, per_page):
                fut = tp.submit(self._request_issues, query=query, start=start, per_page=per_page)
                futures.append(fut)
            try:
                for i, fut in enumerate(concurrent.futures.as_completed(futures), 1):
                    data = fut.result()
                    items = data["item"]
                    if not isinstance(items, list):
                        items = [items]
                    for item in items:
                        for field in self.DATE_FIELDS:
                            item[field] = self._string_to_date(item[field])
                        yield item, total_results
                        total += 1
                        if total > limit:
                            return
            except KeyboardInterrupt:
                cancelled = 0
                for fut in futures:
                    if fut.cancel():
                        cancelled += 1
                print("Futures canceled {} / {}. Shutting down loop, please wait...".format(cancelled, len(futures)))
                tp.shutdown()

    def _request_issues(self, query="order by updated DESC", start=0, per_page=10):
        resp = self._client.get(
            posixpath.join(self._url, "sr/jira.issueviews:searchrequest-xml/temp/SearchRequest.xml"),
            auth=self._auth,
            params={
                "jqlQuery": query,
                "pager/start": start, "tempMax": per_page,
            }
        )
        if resp.status_code != 200:
            raise Exception(resp.content)
        parsed = self.parse_jira_xml(resp.content)
        result = parsed["rss"]["channel"]
        return result

    def parse_jira_xml(self, content):
        return xmltodict.parse(content, dict_constructor=dict)

    def _string_to_date(self, s):
        return datetime.datetime.strptime(s, self.DATE_FORMAT).replace(tzinfo=None)



import traceback

import html2text
from whoosh.fields import Schema, TEXT, KEYWORD, ID
from whoosh.analysis import LanguageAnalyzer
from whoosh.filedb.filestore import FileStorage
from whoosh.highlight import WholeFragmenter
from whoosh.qparser import MultifieldParser

from .issue import Issue
from .utils import ansi_highlight, AnsiColorFormatter
from .downloader import JiraDownloader
from .db import Ticket, db


class JiraSearcher:

    _per_page = 50

    def __init__(self, index_dir="index_dir"):
        self._index_dir = index_dir
        self._html_to_text = html2text.HTML2Text(bodywidth=0)
        self._html_to_text.ignore_links = True
        self._html_to_text.ignore_images = True

    def update_db(self, downloader: JiraDownloader, projects=None, all_issues=False, optimize=False):
        self.report("Updating issues info in {}...".format(projects or "All projects"))
        ix = self._get_index()
        writer = ix.writer()
        with db.atomic():
            count = 0
            if all_issues:
                min_date = max_date = None
            else:
                min_date, max_date = Ticket.get_min_max_dates()
            issues = downloader.iter_issues(
                projects=projects, per_page=self._per_page, min_date=min_date, max_date=max_date
            )
            for i, (issue, total) in enumerate(issues):
                issue["id"] = issue["key"]["#text"]

                Ticket.insert(
                    key=issue["id"], updated=issue["updated"], created=issue["created"],
                    data=issue,
                ).on_conflict_replace().execute()
                try:
                    doc = Issue(issue).to_doc()
                except Exception as exc:
                    self.report("Error while processing {}: {}\n{}".format(
                        issue.get("id"), traceback.format_exc(), issue
                    ))
                    continue
                writer.update_document(**doc)
                self.report(" {} / {} - {} ({}) synced".format(i, total, issue["id"], issue["updated"]))
                if i and i % self._per_page == 0:
                    self.report("{} / {} items downloaded and indexed".format(i, total))
                count += 1
            db.commit()
            if count:
                self.report("Saving index{}...".format(" with optimization" if optimize else ""))
                writer.commit(optimize=optimize)

    def report(self, *args, **kwargs):
        print(*args, **kwargs)

    def _get_schema(self):
        lang_analyzer = LanguageAnalyzer("ru")
        return Schema(
            key=ID(stored=True, unique=True),
            assignee=ID(stored=True),
            reporter=ID(stored=True),
            status=ID(stored=True),
            summary=TEXT(analyzer=lang_analyzer, field_boost=2.0),
            description=TEXT(analyzer=lang_analyzer),
            comments_str=TEXT(analyzer=lang_analyzer),
            labels=KEYWORD(stored=True, lowercase=True),
            components=KEYWORD(stored=True, lowercase=True),
        )

    def _get_index(self):
        storage = FileStorage(self._index_dir).create()
        if storage.index_exists():
            ix = storage.open_index()
        else:
            ix = storage.create_index(self._get_schema())
        return ix

    def search(self, query_str, limit=30, html=True, description=True, comments=False, search_comments=True,
               highlight=True):
        index = self._get_index()
        searcher = index.searcher()
        fields = ["summary", "description"]
        if search_comments:
            fields.append("comments_str")
        qp = MultifieldParser(fields, schema=index.schema)
        query = qp.parse(query_str)

        results = searcher.search(query, limit=limit)
        results.formatter = AnsiColorFormatter()
        results.fragmenter = WholeFragmenter()
        self.report(results)

        for hit in results:
            ticket = Ticket.get_by_id(hit["key"])
            text = Issue(ticket.data).to_string(with_description=description, with_comments=comments)
            if not html:
                text = self._html_to_text.handle(text)
            text = text.strip()
            if highlight and description:
                highlighted = hit.highlights("description", text=text)
                if highlighted:
                    text = highlighted
            self.report(text)
            if description or comments:
                self.report("-" * 80)

    def dump(self, html=False, comments=True, description=True, regex=None, regex_group_name="word"):
        for ticket in Ticket.query_reverse_alphanum():
            issue = Issue(ticket.data)
            text = issue.to_string(with_description=description, with_comments=comments)
            if not html:
                text = self._html_to_text.handle(text)
            if regex:
                text, count = regex.subn(ansi_highlight(r"\g<{}>".format(regex_group_name)), text)
                if not count:
                    continue
            self.report(text.strip())
            if description or comments:
                self.report("-" * 80)



from peewee import Model, SqliteDatabase, CharField, DateTimeField, fn

from .fields import JSONGzipField


db = SqliteDatabase(None)


class BaseModel(Model):
    class Meta:
        database = db


class Ticket(BaseModel):
    key = CharField(primary_key=True)
    updated = DateTimeField(index=True)
    created = DateTimeField(index=True)
    data = JSONGzipField()

    @classmethod
    def get_min_max_dates(cls):
        return cls.select(fn.Min(cls.updated), fn.Max(cls.updated)).scalar(as_tuple=True)

    @classmethod
    def query_reverse_alphanum(cls):
        return cls.select(cls).order_by(fn.Length(cls.key), cls.key).namedtuples().iterator()


def connect(path):
    db.init(path)
    db.connect()
    db.create_tables([Ticket])



import os
import re
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

install_requires = [
    "html2text>=2018.1.9",
    "requests>=2.21.0",
    "Whoosh>=2.7.4",
    "xmltodict>=0.12.0",
    "peewee>=3.9.3",
]


def read(*parts):
    with open(os.path.join(here, *parts), "r") as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name="jirafts",
    url="https://github.com/reclosedev/jirafts/",
    version=find_version("jirafts", "__init__.py"),
    author="Roman Haritonov",
    description="Tool for searching in local copy of JIRA issues",
    author_email="reclosedev@gmail.com",
    license="MIT",
    packages=find_packages("."),
    include_package_data=True,
    install_requires=install_requires,
    entry_points={
        "console_scripts":
            [
                "jirafts = jirafts.cli:main",
            ]
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
)
